open Gui

(********** definition of two level types **********)
(********** ファンクターの引数の型 **********)
module type STRUCTURE_SIG =
sig
  (* ユーザ定義の型 *)
  type 'a t
    (* map 関数 *)
  val map : ('a -> 'b) -> 'a t -> 'b t
    (* acc 関数 : 中身を展開し、集める関数 *)
    (*  val acc : 'a t -> ('a -> 'a -> 'a) -> 'a ->'a (* 引数, 関数, 初期値 *) *)
    (* 単一化関数 *)
  val unify : ('a -> 'a -> unit) -> 'a t -> 'a t -> unit
    (* 描画関数 *)
  val draw : id_t t -> id_t
    (* search関数 *)
  val search : bool t -> bool
end
  
(********** ファンクターの定義 **********)
module F (S : STRUCTURE_SIG) : 
  (sig
   type fix_t = 
       Fix of fix_t S.t * id_t list ref
     | Meta of string *	                (* メタ変数名 *)
	 fix_t option ref *	(* 実体へのポインタ *)
	       id_t list ref
     | Code of (unit -> id_t) * code_t ref * id_t list ref
   and code_t = 
       C of (unit -> fix_t)
     | V of fix_t
   val resubst : fix_t -> fix_t -> bool 
   val get_id : fix_t -> id_t list
   val take : fix_t -> fix_t S.t
   val occur : string -> fix_t -> bool
   val unify : fix_t -> fix_t -> unit
   val draw : fix_t -> id_t (* id_t を返すように変更 *)
   val clear : fix_t -> bool
   val make_meta : string -> fix_t
   val make_fix : fix_t S.t -> fix_t
   val make_code : (unit -> id_t) -> (unit -> fix_t) -> fix_t
   val init : unit -> unit
   val deref : fix_t -> fix_t
   val search : fix_t -> bool
     (* parser登録 *)
   val parse : (string -> fix_t) ref
end)
= 
struct
  (* 型定義 *)
  type fix_t = 
      Fix of fix_t S.t * id_t list ref 
    | Meta of string * fix_t option ref * id_t list ref
    | Code of (unit -> id_t) * code_t ref * id_t list ref
   and code_t = 
       C of (unit -> fix_t)
     | V of fix_t

  (* parser を持たせる *)
  let parse = ref (fun str -> raise Not_found)

  (* メタ変数を生成する *)
  let num = ref 0
  let make_meta str = 
    num := !num + 1;
    Meta(str^(string_of_int !num), ref None, ref [])

 (* ref の一番先を取ってくる : fix_t -> fix_t
     この時、Codeは実行しておく*)
  let rec deref t = match t with
      Meta(_, {contents = Some(t')} , _) -> deref t'
    | Code(str1, c, _) ->
	(match !c with
	     C(f) -> let res = f () in
	       (match res with
		    Code(str2, _, _) -> 
		      c := V(res); res
		      (*
		      if str1 = str2
		      then t
		      else (let res = deref res in c := V(res); res)
		      *)
		  | Meta _ | Fix _ -> 
		      let res = deref res in
			c := V(deref res); res)
	   | V(v) -> 
	       let v = deref v in
		 c := V(v); v)
    | _ -> t


 (* id_t を取り出す : fix_t -> id_t list *) (* 不要 *)
  let rec get_id t = match t with
      Fix (_, id) | Code(_, _, id) | Meta(_, _, id) -> !id

  (* Fix を生成する *)
  let make_fix elt = Fix(elt, ref [])

  (* Codeを生成する *)
  let make_code str f = Code(str, ref (C(f)), ref []) 

  exception Cannot_Change

  (* resubst : Undo の時に使おうと思っている *)
  let rec resubst fix1 fix2 = (*match deref fix1 with
      Meta(_, ({contents = None} as op), _) ->
	op := Some(fix2); true
    | Fix(tm, _) -> S.search (S.map (fun fix1 -> resubst fix1 fix2) tm)
    | _ -> false*) false (* あとで *)
    
  let subst fix =
    let rec subst fix = match fix with
	(* 自分がオリジナルのケース *)
	Meta(_, ({contents = None} as op), _) -> 
	  op := Some(!parse (!input_str^";;"))
(*      | (* 自分はコピーのケース *)
	  Meta(_, {contents = Some (fix')}, _, _) -> subst fix' *)
      | _ -> assert false
    in input_dialog(); subst fix

(* id_t list を全て空リストにする *)
  let rec clear fix = match fix with
     Fix (tm, ids) ->
       ids := [];
       let tm' = S.map clear tm in
	 S.search tm'
    | Meta (_, {contents = None}, ids) ->
	(ids := []; false)
    | Meta (_, {contents = Some (fix')}, ids) ->
	ids := [];
	clear fix';
    | Code (draw_fun, code, ids) ->
	ids := [];
	(match !code with
	     C _ -> false
	   | V(f) -> clear f)

  (* 置換関数 *)
  let rec search fix = 
    match !mode with
	SUB -> (
	  match fix with
	      Meta(_, {contents = None}, ids) ->
		if create_bbox !ids
		then (subst fix; true)
		else false
	    | Meta(_, {contents = Some(fix2)}, ids) -> search fix2
	    | Fix(tm, id) -> S.search (S.map (fun t -> search t) tm)
	    | Code(_, c, _) -> (match !c with
				    C _ -> false
				  | V(v) -> search v))
      | OCC -> false
      | CLEAR -> clear fix

  (* Fixの中身を取ってくる *)
  let rec take fix = match deref fix with
      Fix(tm, _) -> tm
    | _ -> raise Cannot_take 

 let rec occur str fix = 
   match fix with
       Fix(term, _) -> S.search (S.map (occur str) term)
     | Meta(str', _, _) -> str = str'
     | Code(_, _, _) -> true

  (* unification : fix_t -> fix_t -> unit *)
  let rec unify fix1 fix2 = match (deref fix1, deref fix2) with
    | (Meta(_, op1, _), (Meta(_, op2, _) as fix2')) ->
	if op1 != op2 
	then op1 := Some(fix2')
    | (Meta(str, op, _), ((Fix _) as v)) 
    | (((Fix _) as v), Meta(str, op, _)) ->
	if occur str v
	then raise (Unify_Error str)
	else op := Some (v)
    | ((Meta(_, op, _) as m), Code(_, c, _))
    | (Code  (_, c, _), (Meta (_, op, _) as m)) -> 
	(match !c with 
	     C _ -> raise (Unify_Error("of_meta_and_code"))
	   | V(v) -> unify m v)
    | (Fix(t1, _), Fix(t2, _)) -> S.unify unify t1 t2
    | (Code _, Code _) -> ()
    | (Code _, Fix _) | (Fix _, Code _) 
	-> raise (Unify_Error("of_fix_and_code"))

  (* 描画関数 : fix_t -> id_t *)
  let rec draw fix = match fix with
      Fix (tm, ids) ->
	let tm' = S.map draw tm in
	let id = S.draw tm' in
	ids := id :: !ids;
	id
    | Meta (str, {contents = None}, ids) ->
	let id = create_str ~color:meta_color str in
	ids := id :: !ids;
	id
    | Meta (str, {contents = Some (fix')}, ids) ->
	let id = draw fix' in
	ids := id :: !ids;
	id
    | Code (draw_fun, code, ids) ->
	(match !code with
	     C _ -> let id = draw_fun ()
			(* combineH[create_str ~color:code_color str] *) in
		    ids := id :: !ids;
		    id
	   | V(f) -> let id = draw f in
		     ids := id :: !ids;
		     id)	    
	      
  (* 初期化関数 *)
  let init() = num := 0
end
