open Tk
exception Error
exception Error of string
exception Cannot_take
exception Unify_Error of string

(********** types for gui objects **********)
(* object identifire *)
type id_t = ID of Tk.tagOrId | NEW

(* to get a tagOrId *)
let get_tag = function 
    ID(t) -> t
  | NEW -> raise (Protocol.TkError "No identifire.")

(********** make a main frame **********)
(* make a top window *)     
let top = openTk () 

(* a frame for a canvas to scroll *)
let frame_cv = Frame.create top
  
(* the color of a canvas *)
let canvas_color = ref (`White)

(* make a canvas *)
let canvas = 
  Canvas.create frame_cv ~background:!canvas_color
    ~width:1000 ~height:500 ~scrollregion:(0, 0, 1000, 500)
    (*~height:200 ~width:400*)

(* input dialog *)
let input_str = ref "" (* 文字列は、このままで良い？ *)
let input_dialog () = 
  let top2 = Toplevel.create top in
  let message = Label.create top2 
    ~text:"Enter a substitution term." ~width:20 in
  let ent = Entry.create top2 ~width:20 in
  let qbutton = Button.create ~takefocus:false ~text:"OK" top2
    ~command:(fun () -> input_str := Entry.get ent;
		destroy top2) in
    pack [coe message; coe ent; coe qbutton];
    Tkwait.window top2; 
    input_str := !input_str ^ ";;"

(* a frame to arrange buttons user define *)
let user_frame = Frame.create top

(* a buttom frame *)
let frame_bottom = Frame.create top 

(* a button to quit *)
let b_quit = Button.create ~takefocus:false ~text:"Quit" frame_bottom 
    ~command:(closeTk) 

(* create a new button on the buttom frame *)
let create_button_low text func = 
  pack ~ipadx:5 ~padx:5 ~side:`Left 
    [Button.create ~takefocus:false ~command:func ~text:text frame_bottom]

(* the window to inform all inferences are completed *)
let show_comp () = 
  ignore(Tk.messageBox ~default:"ok" ~icon:`Info 
	   ~message:"Complete!!" ~parent:top ~title:"Congraturations!" ~typ:`Ok ())
    
(* display a select dialog *)
let create_dialog title message text_lst =
  Dialog.create ~parent:canvas ~title:title ~message:message 
    ~buttons:text_lst ~default:0 () 

(********** variables **********)
(* initial value *)
let init_x = 100
let init_y = 400

(* coords for drawing and mouse pointer *)
let coord_x = ref init_x
let coord_y = ref init_y

(* the vertical margin *)
let sep = 2

(* font *)
let font = ref "Century 15"

(* color *)
let font_color = ref (`Black)
let meta_color = `Color("#555555")
let code_color = `Color("#222288")
let gray = `Color("#777777")

(* save a pointed term *)
let pointed_id = ref NEW (* dummy *)

(* change the scroll region *)
let change_region () = 
  let w = Winfo.width canvas in
  let h = Winfo.height canvas in
    let (x1, y1, x2, y2) = (try Canvas.bbox canvas [`Tag("all")] with 
			  Protocol.TkError _ -> (0,0,w,h)) in
    if (x2 - x1) < w || (y2 - y1) < h
    then Canvas.configure ~scrollregion:(x1 - 50, y1 - 50, x2 + 50, y2 + 50) canvas

(* initialize function *)
let init() = coord_x := init_x; coord_y := init_y; pointed_id := NEW

(********** functions used by users **********)
(* drawing function : string list -> id_t *)
let create_str ?(color = !font_color) str =
	let tag = Canvas.create_text ~fill:color  ~font:!font ~x:!coord_x ~y:!coord_y 
	  ~text:str ~anchor:`Sw canvas 
	in ID(tag)
   
(* generate an original tag *)
let tag_num = ref 0
let add() = tag_num := !tag_num + 1
let make_tag () = add (); "a" ^ (string_of_int !tag_num)

(* arrange items horizontally (from left to right) : id_t list -> int -> id_t *)
let combine_des idlst par = 
  let taglst = List.map get_tag idlst in
  let tag = make_tag () in
  let _ = List.fold_left (fun x t ->
			    Canvas.addtag canvas ~tag:tag ~specs:[`Withtag(t)];
			    let (x1,_,x2,_) = Canvas.bbox canvas [t] in
			      Canvas.move canvas t ~x:x ~y:0;
			      x + (x2 - x1) + par) 0 taglst in ID(`Tag(tag))

(* arrange two items horizontally *)
let combineH idlst = combine_des idlst 0

(* arrange items vertically (from bottom to top) *)
let puts ob_lst = 
  let tag_lst = List.map get_tag ob_lst in
  let tag = make_tag () in
  let _ = List.fold_left
    (fun y t ->
       Canvas.addtag canvas ~tag:tag ~specs:[`Withtag(t)];
       let (_,y1,_,y2) = Canvas.bbox canvas [t] in
	 Canvas.move canvas t ~x:0 ~y:y;
	 y + (y1 - y2) - sep) 0 tag_lst in ID(`Tag(tag))

(* arrange two items vertically *)
let combineV up_id under_id =
	let (x1,_,x2,_) = Canvas.bbox canvas [get_tag up_id; get_tag under_id] in
	let line = ID(Canvas.create_line ~xys:[(!coord_x,!coord_y);
					       (!coord_x+(x2-x1),!coord_y)] canvas) in
	  puts [up_id; line; under_id]

(* arrange two items vertically with middle line *)
let put_rule up_id under_id str = 
  let tree_id = get_tag (combineV up_id under_id) in
  let (x1,y1,x2,y2) = Canvas.bbox canvas [tree_id] in
  let tag = make_tag () in
  let _ = Canvas.create_text ~fill:!font_color ~font:"Century 10" ~tags:[tag] 
    ~x:(x2+2) ~y:((y1 + y2)/2+8) ~text:("("^str^")") ~anchor:`Sw canvas in
    Canvas.addtag canvas ~tag:tag ~specs:[`Withtag(`Tag(tag))];
    ID(`Tag(tag))

(* arrange an axiom *)
let combineV_axm id =
  let (x1,_,x2,_) = Canvas.bbox canvas [get_tag id] in
  let line = ID(Canvas.create_line 
		  ~xys:[(!coord_x, !coord_y); (!coord_x+(x2-x1), !coord_y)] canvas) in
    puts [id; line]
           
(* delete bounding boxes *)
let delete_bbox () = Canvas.delete canvas [`Tag("bbox")]
  
(* check if an item is pointed *)
let check id =
  if id = NEW
  then raise (Protocol.TkError "error of identifire") (* for debug *)
  else 
    let t = get_tag id in
    let (x1,y1,x2,y2) = Canvas.bbox canvas [t] in
      x1 <= !coord_x && !coord_x <= x2 && y1 <= !coord_y && !coord_y <= y2
	
(* display a bounding box if the item is pointed *)
let create_bbox_1 id = 
  if check id
  then 
    ( let t = get_tag id in
      let (x1,y1,x2,y2) = Canvas.bbox canvas [t] in
      let _ = Canvas.create_rectangle ~x1:x1 ~y1:y1 ~x2:x2 ~y2:y2 ~tags:["bbox"]
	~outline:`Blue canvas in 
	pointed_id := id; true)
  else false

let rec create_bbox id_list = match id_list with
    [] -> false
  | id :: rest -> create_bbox_1 id || create_bbox rest

(* たぶんいらない関数 *)
let create_bbox2 id =
  if id = NEW then (print_string "dame"; print_newline(); false)
  else 
    (print_string "OK!"; print_newline();
  let t = get_tag id in
  let (x1,y1,x2,y2) = Canvas.bbox canvas [t] in
  let _ = Canvas.create_rectangle ~x1:x1 ~y1:y1 ~x2:x2 ~y2:y2 ~tags:["bbox"]
    ~outline:`Blue canvas in 
    true)
      
(* make a new button (text:string, f: unit -> unit) *)
let create_button text f = 
  pack ~side:`Left [Button.create ~takefocus:false
		      ~command:(fun _ -> f() ; change_region())  ~text:text user_frame]

(* display parenthes if they are needed *)
let need_paren need id =
  if need 
  then 
    let op = create_str "(" in
    let cl = create_str ")" in
      combineH [op; id; cl]
  else id

(* idが入っているかを判定する *) (* いらない？ *)
let judge_inside id = 
  let (x1, y1, x2, y2) = Canvas.bbox canvas [get_tag id] in
  let (x1', y1', x2', y2') = Canvas.bbox canvas [get_tag (!pointed_id)] in
    x1 <= x1' && x2' <= x2 && y1 <= y1' && y2' <= y2

(* a type user neeed to define *)
type 'a prule_t = 
    Infer of 'a * 'a prule_t list
  | Axiom of 'a
  
type mode_t = OCC | SUB | CLEAR
let mode = ref OCC
